/*
 * Copyright (c) 2001 Edouard TISSERANT <tissered@esstin.u-nancy.fr>
 *  Parts inspired from Shane Watts <shane@bofh.asn.au> Xfree 3 Acecad Driver
 *  Thanks to Emily, from AceCad, For giving me documents.
 * Copyright (c) 2010 Gerard Krol <gerard_@gna.org>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
 * OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 *
 */

#include "config.h"

#define _WIZARDPEN_C_
/*****************************************************************************
 *  Standard Headers
 ****************************************************************************/

#ifdef LINUX_INPUT
    #include <asm/types.h>
    #include <linux/input.h>
    #ifndef EV_SYN
        #define EV_SYN EV_RST
        #define SYN_REPORT 0
    #endif
    #ifdef BUS_PCI
        #undef BUS_PCI
    #endif
    #ifdef BUS_ISA
        #undef BUS_ISA
    #endif
#endif

#include <misc.h>
#include <xf86.h>

#ifndef NEED_XF86_TYPES
    #define NEED_XF86_TYPES
#endif
#include <xf86_OSproc.h>
#include <xisb.h>
#include <xf86Xinput.h>
#include <exevents.h>
#include <xf86Module.h>

#include <string.h>
#include <stdio.h>

#ifdef LINUX_INPUT
    #include <errno.h>
    #include <fcntl.h>
    #ifdef LINUX_SYSFS
        #include <sysfs/libsysfs.h>
        #include <dlfcn.h>
    #endif
#endif

/*****************************************************************************
 *  Local Headers
 ****************************************************************************/
#include "wizardpen.h"

/*****************************************************************************
 *  Variables without includable headers
 ****************************************************************************/

/*****************************************************************************
 *  Local Variables
 ****************************************************************************/

#undef read
#define read(a,b,c) xf86ReadSerial((a),(b),(c))

/* max number of input events to read in one read call */
#define MAX_EVENTS 50

static const char *default_options[] =
{
    "BaudRate", "9600",
    "StopBits", "1",
    "DataBits", "8",
    "Parity", "Odd",
    "Vmin", "1",
    "Vtime", "10",
    "FlowControl", "Xoff",
    NULL
};

_X_EXPORT InputDriverRec WIZARDPEN =
{
    6,
    "wizardpen",
    NULL,
    WizardPenPreInit,
    NULL,
    NULL,
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
    0
#else
    default_options
#endif
};

#ifdef XFree86LOADER
static XF86ModuleVersionInfo VersionRec =
{
    "wizardpen",
    MODULEVENDORSTRING,
    MODINFOSTRING1,
    MODINFOSTRING2,
    XORG_VERSION_CURRENT,
    PACKAGE_VERSION_MAJOR, PACKAGE_VERSION_MINOR, PACKAGE_VERSION_PATCHLEVEL,
    ABI_CLASS_XINPUT,
    ABI_XINPUT_VERSION,
    MOD_CLASS_XINPUT,
    {0, 0, 0, 0}
};


_X_EXPORT XF86ModuleData wizardpenModuleData = {
    &VersionRec,
    SetupProc,
    TearDownProc
};

/*****************************************************************************
 *  Function Definitions
 ****************************************************************************/

static pointer
SetupProc(  pointer module,
        pointer options,
        int *errmaj,
        int *errmin )
{
    xf86AddInputDriver(&WIZARDPEN, module, 0);
    return module;
}

static void
TearDownProc( pointer p )
{
}
#endif // #ifdef XFree86LOADER

#ifdef LINUX_INPUT
static int
IsUSBLine(int fd)
{
    int version;
    int err;

    SYSCALL(err = ioctl(fd, EVIOCGVERSION, &version));

    if (!err) {
        xf86MsgVerb(X_PROBED, 4, "Kernel Input driver version is %d.%d.%d\n",
                version >> 16, (version >> 8) & 0xff, version & 0xff);
        return 1;
    } else {
        xf86MsgVerb(X_PROBED, 4, "No Kernel Input driver found\n");
        return 0;
    }
}

/* Heavily inspired by synpatics/eventcomm.c */

#define DEV_INPUT_EVENT "/dev/input/event"
#define EV_DEV_NAME_MAXLEN 64
#define SET_EVENT_NUM(str, num) \
    snprintf(str, EV_DEV_NAME_MAXLEN, "%s%d", DEV_INPUT_EVENT, num)

static Bool
fd_query_wizardpen(int fd, char *wizardpen_name) {
    char name[256] = "Unknown";
    int cmp_at = strlen(wizardpen_name);
    if (cmp_at > 255)
        cmp_at = 255;
    ioctl(fd, EVIOCGNAME(sizeof(name)), name);
    name[cmp_at] = '\0';
     if (xf86NameCmp(name, wizardpen_name) == 0)
        return TRUE;
    return FALSE;
}

static char wizardpen_name_default[10] = "  TABL";

static Bool
WizardPenAutoDevProbe(InputInfoPtr local, int verb)
{
    /* We are trying to find the right eventX device */
    int i = 0;
    Bool have_evdev = FALSE;
    int noent_cnt = 0;
    const int max_skip = 10;
    char *wizardpen_name = xf86FindOptionValue(local->options, "Name");
    char fname[EV_DEV_NAME_MAXLEN];
    int np;

#ifdef LINUX_SYSFS
    struct sysfs_bus *usb_bus = NULL;
    struct sysfs_driver *wizardpen_driver = NULL;
    struct sysfs_device *candidate = NULL;
    char *link = NULL;
    struct dlist *devs = NULL;
    struct dlist *links = NULL;
    unsigned int major = 0, minor = 0;
    void *libsysfs = NULL;
#endif

    if (!wizardpen_name)
        wizardpen_name = wizardpen_name_default;

    xf86MsgVerb(X_INFO, verb, "%s: probing event devices for WizardPen tablets\n", local->name);
    for (i = 0; ; i++) {
        int fd = -1;
          Bool is_wizardpen;

        np = SET_EVENT_NUM(fname, i);
        if (np < 0 || np >= EV_DEV_NAME_MAXLEN) {
            xf86MsgVerb(X_WARNING, verb, "%s: too many devices, giving up %d\n", local->name, i);
            break;
        }
        SYSCALL(fd = open(fname, O_RDONLY));
        if (fd < 0) {
            if (errno == ENOENT) {
                if (++noent_cnt >= max_skip)
                    break;
                else
                    continue;
            } else {
                continue;
            }
        }
        noent_cnt = 0;
        have_evdev = TRUE;
        is_wizardpen = fd_query_wizardpen(fd, wizardpen_name);
        SYSCALL(close(fd));
        if (is_wizardpen) {
            goto ProbeFound;
        }
    }
    xf86MsgVerb(X_WARNING, verb, "%s: no WizardPen event device found (checked %d nodes, no device name started with '%s')\n",
            local->name, i + 1, wizardpen_name);
    if (i <= max_skip)
        xf86MsgVerb(X_WARNING, verb, "%s: The /dev/input/event* device nodes seem to be missing\n",
                local->name);
    if (i > max_skip && !have_evdev)
        xf86MsgVerb(X_WARNING, verb, "%s: The evdev kernel module seems to be missing\n", local->name);
    return FALSE;

ProbeFound:
    xf86Msg(X_PROBED, "%s auto-dev sets device to %s\n",
            local->name, fname);
    xf86ReplaceStrOption(local->options, "Device", fname);
    return TRUE;
}

#endif // #ifdef LINUX_INPUT

static int parsePercentage(const char *s, int max)
{
    const char *p;
    int v = atoi(s);
    if (!*s)
        return 0;
    // does it end on '%'?
    for (p=s;*p;p++);
    if (*(p-1) == '%')
    {
        return ((int64_t)v * max)/100;
    }
    // no? just the value then
    return v;
}

// FIXME: use the standard config parsing functions
static BOOL parseBool(const char *s)
{
    if (!*s)
        return FALSE;
    if (atoi(s) > 0)
        return TRUE;
    return strcmp(s, "yes") == 0
        || strcmp(s, "on") == 0
        || strcmp(s, "true") == 0;
}

#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
static int
NewWizardPenPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags);

static InputInfoPtr
WizardPenPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
{
    LocalDevicePtr local = xf86AllocateInput(drv, 0);

    if (!local)
        goto SetupProc_fail;

    local->name = dev->identifier;
    local->flags = XI86_SEND_DRAG_EVENTS;
    local->conf_idev = dev;

    if (NewWizardPenPreInit(drv, local, flags) == Success)
    {
        local->flags |= XI86_CONFIGURED;
        return local;
    }

    xf86DeleteInput(local, 0);
SetupProc_fail:
    return NULL;
}

static int
NewWizardPenPreInit(InputDriverPtr drv, InputInfoPtr local, int flags)
#else
static int
WizardPenPreInit(InputDriverPtr drv, InputInfoPtr local, int flags)
#endif
{
    WizardPenPrivatePtr priv = calloc (1, sizeof(WizardPenPrivateRec));
    int msgtype;
    char *s;
    int rc = Success;

    if (!priv) {
        rc = BadAlloc;
        goto SetupProc_fail;
    }

    local->type_name = "WizardPen Tablet";
    local->private = priv;
    local->device_control = DeviceControl;
    /*local->always_core_feedback = 0;*/
    xf86CollectInputOptions(local, default_options
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
        , NULL
#endif
    );

    xf86OptionListReport(local->options);

    priv->increment = xf86SetIntOption(local->options, "Increment", 0 );

    /* deprecated settings */
    s = xf86FindOptionValue(local->options, "ClickPressureLevel");
    if (s && strlen(s) > 0) {
        xf86Msg(X_WARNING, "%s: Option \"ClickPressureLevel\" "
            "is DEPRECATED, please use \"TopZ\"\n", local->name);
        priv->topZ = atoi(s);
        xf86Msg(X_CONFIG, "%s: TopZ = %d\n", local->name, priv->topZ);
    }
    s = xf86FindOptionValue(local->options, "debugyn");
    if (s && strlen(s) > 0) {
        xf86Msg(X_WARNING, "%s: Option \"debugyn\" "
            "is DEPRECATED, please use \"DebugLevel\"\n", local->name);
        priv->debugLevel = atoi(s);
        xf86Msg(X_CONFIG, "%s: DebugLevel = %d\n", local->name, priv->debugLevel);
    }
    /* end of deprecated settings */

    // increment setting
    priv->increment = xf86SetIntOption(local->options, "Increment", 0);
    s = xf86FindOptionValue(local->options, "Device");
    if (!s || (s && (xf86NameCmp(s, "auto-dev") == 0))) {
#ifdef LINUX_INPUT
        priv->flags |= AUTODEV_FLAG;
        if (!WizardPenAutoDevProbe(local, 0))
        {
            xf86Msg(X_ERROR, "%s: unable to find device\n", local->name);
            goto SetupProc_fail;
        }
#else
        xf86Msg(X_NOT_IMPLEMENTED, "%s: device autodetection not implemented, sorry\n", local->name);
        goto SetupProc_fail;
#endif
    }

    local->fd = xf86OpenSerial (local->options);
    if (local->fd == -1)
    {
        xf86Msg(X_ERROR, "%s: unable to open device\n", local->name);
        goto SetupProc_fail;
    }
    xf86ErrorFVerb( 6, "tty port opened successfully\n" );

#ifdef LINUX_INPUT
    if (IsUSBLine(local->fd)) {
        priv->flags |= USB_FLAG;

        local->read_input = USBReadInput;

        if (USBQueryHardware(local) != Success)
        {
            xf86Msg(X_ERROR, "%s: unable to query/initialize hardware (not an %s?).\n", local->name, local->type_name);
            goto SetupProc_fail;
        }
    } else
#endif

    s = xf86FindOptionValue(local->options, "Mode");
    msgtype = s ? X_CONFIG : X_DEFAULT;
    if (!(s && (xf86NameCmp(s, "relative") == 0)))
    {
        priv->flags |= ABSOLUTE_FLAG;
    }

    xf86Msg(msgtype, "%s is in %s mode\n", local->name, (priv->flags & ABSOLUTE_FLAG) ? "absolute" : "relative");

    /* box settings */
    s = xf86FindOptionValue(local->options, "TopX");
    if (!s || !*s)
    {
        s = "5%";
        xf86Msg(X_CONFIG, "%s: TopX not set, defaulting to \"%s\"\n", local->name, s);
    }
    priv->topX = parsePercentage(s, priv->maxX);

    s = xf86FindOptionValue(local->options, "TopY");
    if (!s || !*s)
    {
        s = "5%";
        xf86Msg(X_CONFIG, "%s: TopY not set, defaulting to \"%s\"\n", local->name, s);
    }
    priv->topY = parsePercentage(s, priv->maxY);


    s = xf86FindOptionValue(local->options, "BottomX");
    if (!s || !*s)
    {
        s = "95%";
        xf86Msg(X_CONFIG, "%s: BottomX not set, defaulting to \"%s\"\n", local->name, s);
    }
    priv->bottomX = parsePercentage(s, priv->maxX);

    s = xf86FindOptionValue(local->options, "BottomY");
    if (!s || !*s)
    {
        s = "95%";
        xf86Msg(X_CONFIG, "%s: BottomY not set, defaulting to \"%s\"\n", local->name, s);
    }
    priv->bottomY = parsePercentage(s, priv->maxY);

    s = xf86FindOptionValue(local->options, "TopZ");
    if (s && strlen(s) > 0) {
        priv->topZ = parsePercentage(s, priv->maxZ);
    }
    s = xf86FindOptionValue(local->options, "BottomZ");
    if (s && strlen(s) > 0) {
        priv->bottomZ = parsePercentage(s, priv->maxZ);
    }
    
    s = xf86FindOptionValue(local->options, "KeepShape");
    if (s && strlen(s) > 0) {
        priv->keepShape = parseBool(s);
        xf86Msg(X_CONFIG, "%s: KeepShape = %s\n", local->name, priv->keepShape ? "on" : "off");
    }

    s = xf86FindOptionValue(local->options, "ScreenX");
    if (s && strlen(s) > 0) {
        priv->screenX = parsePercentage(s, priv->screenX);
        xf86Msg(X_CONFIG, "%s: ScreenX = %d\n", local->name, priv->screenX);
    }else
        priv->screenX = screenInfo.screens[0]->width;

    s = xf86FindOptionValue(local->options, "ScreenY");
    if (s && strlen(s) > 0) {
        priv->screenY = parsePercentage(s, priv->screenY);
        xf86Msg(X_CONFIG, "%s: ScreenY = %d\n", local->name, priv->screenY);
    }else
        priv->screenY = screenInfo.screens[0]->height;
    xf86Msg(X_INFO, "%s: ScreenX = %i, ScreenY = %i\n", local->name, priv->screenX, priv->screenY);
    if (priv->keepShape)
    {
        float screenAspect = (float)priv->screenX/priv->screenY;
        xf86Msg(X_INFO, "%s: Screen aspect ratio is %.2f\n", local->name, screenAspect);
        if (screenAspect > priv->aspectRatio)
        {
            // screen is wider than the tablet
            // bottomY should be decreased so sizeX/sizeY == screenAspect
            int sizeX = priv->bottomX - priv->topX;
            int sizeY = sizeX/screenAspect*priv->aspectRatio;
            xf86Msg(X_INFO, "%s: Decreasing BottomY from %i to %i\n", local->name, priv->bottomY, priv->topY + sizeY);
            priv->bottomY = priv->topY + sizeY;
        }
        else
        {
            // screen is taller than the tablet
            // bottomX should be decreased so sizeX/sizeY == screenAspect
            int sizeY = priv->bottomY - priv->topY;
            int sizeX = sizeY*screenAspect/priv->aspectRatio;
            xf86Msg(X_INFO, "%s: Decreasing BottomX from %i to %i\n", local->name, priv->bottomX, priv->topX + sizeX);
            priv->bottomX = priv->topX + sizeX;
        }
    }

    xf86Msg(X_CONFIG, "%s: TopX                   = %d\n", local->name, priv->topX);
    xf86Msg(X_CONFIG, "%s: TopY                   = %d\n", local->name, priv->topY);
    xf86Msg(X_CONFIG, "%s: BottomX                = %d\n", local->name, priv->bottomX);
    xf86Msg(X_CONFIG, "%s: BottomY                = %d\n", local->name, priv->bottomY);
    xf86Msg(X_CONFIG, "%s: TopZ    (min pressure) = %d\n", local->name, priv->topZ);
    xf86Msg(X_CONFIG, "%s: BottomZ (max pressure) = %d\n", local->name, priv->bottomZ);

    s = xf86FindOptionValue(local->options, "DebugLevel");
    if (s && strlen(s) > 0) {
        priv->debugLevel = atoi(s);
        xf86Msg(X_CONFIG, "%s: DebugLevel = %d\n", local->name, priv->debugLevel);
    }

    s = xf86FindOptionValue(local->options, "MouseSpeed");
    if (s && strlen(s) > 0) {
        priv->mouseSpeed = atoi(s);
        xf86Msg(X_CONFIG, "%s: MouseSpeed = %d\n", local->name, priv->mouseSpeed);
    }

    s = xf86FindOptionValue(local->options, "MouseAccel");
    if (s && strlen(s) > 0) {
        priv->mouseAccel = parseBool(s);
        xf86Msg(X_CONFIG, "%s: MouseAccel = %d\n", local->name, priv->mouseAccel);
    }
    
    s = xf86FindOptionValue(local->options, "ReportScreen");
    if (s && strlen(s) > 0) {
        priv->reportScreen = parseBool(s);
        xf86Msg(X_CONFIG, "%s: ReportScreen = %s\n", local->name, priv->reportScreen ? "on" : "off");
    }

    s = xf86FindOptionValue(local->options, "Rotate90");
    if (s && strlen(s) > 0) {
        int rotate;
        priv->rotate90 = parseBool(s);
        xf86Msg(X_CONFIG, "%s: Rotate90 = %d\n", local->name, priv->rotate90);
        if(priv->rotate90){
            int temp;
            temp = priv->bottomY;
            priv->bottomY = priv->bottomX;
            priv->bottomX = temp;
            temp = priv->topY;
            priv->topY = priv->topX;
            priv->topX = temp;
        }
    }

    // Option "TPCButton" "on"|"off"
    s = xf86FindOptionValue(local->options, "TPCButton");
    if (s && strlen(s) > 0) {
        priv->tpcButton = parseBool(s);
        xf86Msg(X_CONFIG, "%s: TPCButton = %s\n", local->name, priv->tpcButton ? "on" : "off");
    }

#if 0
    history_size = xf86SetIntOption(local->options , "HistorySize", 0);
#endif

    xf86ProcessCommonOptions(local, local->options);

    if (local->fd != -1)
    {
        RemoveEnabledDevice (local->fd);
        if (priv->buffer)
        {
            XisbFree(priv->buffer);
            priv->buffer = NULL;
        }
        xf86CloseSerial(local->fd);
    }
    RemoveEnabledDevice (local->fd);
    local->fd = -1;
    return rc;

    /*
     * If something went wrong, cleanup and return NULL
     */
SetupProc_fail:
    if ((local) && (local->fd))
        xf86CloseSerial (local->fd);
    if ((priv) && (priv->buffer))
        XisbFree (priv->buffer);
    if (priv)
        free (priv);
    return rc;
}

static Bool
DeviceControl (DeviceIntPtr dev, int mode)
{
    Bool RetValue;

    switch (mode)
    {
        case DEVICE_INIT:
            DeviceInit(dev);
            RetValue = Success;
            break;
        case DEVICE_ON:
            RetValue = DeviceOn(dev);
            break;
        case DEVICE_OFF:
            RetValue = DeviceOff(dev);
            break;
        case DEVICE_CLOSE:
            RetValue = DeviceClose(dev);
            break;
        default:
            RetValue = BadValue;
    }

    return RetValue;
}

static Bool
DeviceOn (DeviceIntPtr dev)
{
    char buffer[256];
    InputInfoPtr local = (InputInfoPtr) dev->public.devicePrivate;
    WizardPenPrivatePtr priv = (WizardPenPrivatePtr) (local->private);

    xf86MsgVerb(X_INFO, 4, "%s Device On\n", local->name);

    local->fd = xf86OpenSerial(local->options);
    if (local->fd == -1)
    {
        xf86Msg(X_WARNING, "%s: cannot open input device %s: %s\n", local->name, xf86FindOptionValue(local->options, "Device"), strerror(errno));
        priv->flags &= ~AVAIL_FLAG;
#ifdef LINUX_INPUT
        if ((priv->flags & AUTODEV_FLAG) && WizardPenAutoDevProbe(local, 4))
            local->fd = xf86OpenSerial(local->options);
        if (local->fd == -1)
#endif
            return !Success;
    }
    priv->flags |= AVAIL_FLAG;


    if (!(priv->flags & USB_FLAG)) {
        priv->buffer = XisbNew(local->fd, 200);
        if (!priv->buffer)
        {
            xf86CloseSerial(local->fd);
            local->fd = -1;
            return !Success;
        }

        /* Rets qu'a l'envoyer a la tablette */
        sprintf(buffer, "%s%c%c%c%c", wizardpen_initstr, priv->reportSpeed, WIZARDPEN_INCREMENT, 32 + priv->increment, (priv->flags & ABSOLUTE_FLAG)? WIZARDPEN_ABSOLUTE: WIZARDPEN_RELATIVE);
        XisbWrite (priv->buffer, (unsigned char *)buffer, strlen(buffer));
    }

    xf86FlushInput(local->fd);
    xf86AddEnabledDevice (local);
    dev->public.on = TRUE;
    return Success;
}

static Bool
DeviceOff (DeviceIntPtr dev)
{
    InputInfoPtr local = (InputInfoPtr) dev->public.devicePrivate;
    WizardPenPrivatePtr priv = (WizardPenPrivatePtr) (local->private);

    xf86MsgVerb(X_INFO, 4, "%s Device Off\n", local->name);

    if (local->fd != -1)
    {
        RemoveEnabledDevice (local->fd);
        if (priv->buffer)
        {
            XisbFree(priv->buffer);
            priv->buffer = NULL;
        }
        xf86CloseSerial(local->fd);
    }


    xf86RemoveEnabledDevice (local);
    dev->public.on = FALSE;
    return Success;
}

static Bool
DeviceClose (DeviceIntPtr dev)
{
    InputInfoPtr local = (InputInfoPtr) dev->public.devicePrivate;

    xf86MsgVerb(X_INFO, 4, "%s Device Close\n", local->name);

    return Success;
}

static void
ControlProc(DeviceIntPtr dev, PtrCtrl *ctrl)
{
    InputInfoPtr local = (InputInfoPtr) dev->public.devicePrivate;

    xf86MsgVerb(X_INFO, 4, "%s Control Proc\n", local->name);
}

static Bool
DeviceInit (DeviceIntPtr dev)
{
    int rx, ry;
    InputInfoPtr local = (InputInfoPtr) dev->public.devicePrivate;
    WizardPenPrivatePtr priv = (WizardPenPrivatePtr) (local->private);
    priv->oldX = 0;
    priv->oldY = 0;
    priv->oldZ = 0;
    unsigned char map[] = {0, 1, 2, 3, 4, 5};
    xf86MsgVerb(X_INFO, 4, "%s Init\n", local->name);

#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
    Atom label=0;
    Atom atoms[MAX_VALUATORS] = { 0 };
#endif

    /* 3 buttons changed to SIX */
    if (InitButtonClassDeviceStruct(
        dev,
        6,
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
        atoms,
#endif
        map) == FALSE)
    {
        xf86Msg(X_ERROR, "%s: unable to allocate ButtonClassDeviceStruct\n", local->name);
        return !Success;
    }

    if (InitFocusClassDeviceStruct (dev) == FALSE)
    {
        xf86Msg(X_ERROR, "%s: unable to allocate FocusClassDeviceStruct\n", local->name);
        return !Success;
    }

    if (InitPtrFeedbackClassDeviceStruct(dev, ControlProc) == FALSE) {
        xf86Msg(X_ERROR, "%s: unable to init ptr feedback\n", local->name);
        return !Success;
    }


    /* 3 axes */
    if (InitValuatorClassDeviceStruct(dev, 3,
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
        atoms,
#endif
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 3
        xf86GetMotionEvents,
#endif
        GetMotionHistorySize(),
        ((priv->flags & ABSOLUTE_FLAG)? Absolute: Relative)|OutOfProximity) == FALSE)
    {
        xf86Msg(X_ERROR, "%s: unable to allocate ValuatorClassDeviceStruct\n", local->name);
        return !Success;
    }
    else
    {
        int minReportX, minReportY;
        int maxReportX, maxReportY;
        if (priv->reportScreen)
        {
            minReportX = minReportY = 0;
            maxReportX = priv->screenX;
            maxReportY = priv->screenY;
        }
        else
        {
            minReportX = priv->topX;
            minReportY = priv->topY;
            maxReportX = priv->bottomX;
            maxReportY = priv->bottomY;
        }
        InitValuatorAxisStruct(dev,
                0,
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
                label,
#endif
                minReportX,
                maxReportX,
                1000,       /* resolution */
                0,          /* min_res */
                1000        /* max_res */
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
        	, (priv->flags & ABSOLUTE_FLAG)? Absolute: Relative
#endif
		);
        InitValuatorAxisStruct(dev,
                1,
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
                label,
#endif
                minReportY,
                maxReportY,
                1000,       /* resolution */
                0,          /* min_res */
                1000        /* max_res */
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
        	, (priv->flags & ABSOLUTE_FLAG)? Absolute: Relative
#endif
		);
        InitValuatorAxisStruct(dev,
                2,
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
                label,
#endif
                0,          /* min val */
                1023,       /* max val */
                1000,       /* resolution */
                0,          /* min_res */
                1000        /* max_res */
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
        	, (priv->flags & ABSOLUTE_FLAG)? Absolute: Relative
#endif
		);

        if (priv->rotate90) {
            int tempval;
            tempval = priv->screenX;
            priv->screenX = priv->screenY;
            priv->screenY = tempval;
        }
    }

    if (InitProximityClassDeviceStruct (dev) == FALSE)
    {
        xf86Msg(X_ERROR, "%s: unable to allocate ProximityClassDeviceStruct\n", local->name);
        return !Success;
    }

    xf86MotionHistoryAllocate (local);

    /* We now know the screen size, so calculate the increment */
    if (priv->increment > 95)
        priv->increment = 95;
    if (priv->increment < 1)
    {
        /* guess the best increment value given video mode */
        rx = priv->maxX / priv->screenX;
        ry = priv->maxY / priv->screenY;
        if (rx < ry)
            priv->increment = rx;
        else
            priv->increment = ry;
        if (priv->increment < 1)
            priv->increment = 1;
    }

    xf86Msg(X_INFO, "%s Increment: %d\n", local->name, priv->increment);

    return Success;
}

#ifdef LINUX_INPUT
#define set_bit(byte,nb,bit)    (bit ? byte | (1<<nb) : byte & (~(1<<nb)))
static void
USBReadInput (InputInfoPtr local)
{
    int len;
    struct input_event * event;
    char eventbuf[sizeof(struct input_event) * MAX_EVENTS];
    WizardPenPrivatePtr priv = (WizardPenPrivatePtr) (local->private);
    int delta;
    int id;
    int x = priv->oldX;
    int y = priv->oldY;
    int z = priv->oldZ;
    int report_x,report_y, report_z;  /* new report_z var */
    int prox = priv->oldProximity;
    int buttons = priv->oldButtons;
    int report_buttons;
    int is_core_pointer = 0;
    /* Is autodev active? */
    int autodev = priv->flags & AUTODEV_FLAG;
    /* Was the device available last time we checked? */
    int avail = priv->flags & AVAIL_FLAG;

    SYSCALL(len = read(local->fd, eventbuf, sizeof(eventbuf)));

    if (len <= 0) {
        if (avail) {
            xf86Msg(X_ERROR, "%s: error reading device %s: %s\n", local->name, xf86FindOptionValue(local->options, "Device"), strerror(errno));
        }
        if (NOTAVAIL) {
            priv->flags &= ~AVAIL_FLAG;
            if(autodev) {
                if (WizardPenAutoDevProbe(local, 4)) {
                    DeviceOff(local->dev);
                    DeviceOn(local->dev);
                }
            }
        }
        return;
    } else {
        if (!avail) {
           // If the device wasn't available last time we checked
            xf86Msg(X_INFO, "%s: device %s is available again\n", local->name, xf86FindOptionValue(local->options, "Device"));
            priv->flags |= AVAIL_FLAG;
        }
    }

    for (event = (struct input_event *)eventbuf;
         event < (struct input_event *)(eventbuf+len);
         event++)
    {
        switch (event->type) {
            case EV_SYN:// 2.6.x
                if (event->code != SYN_REPORT && event->code != 0)
                    xf86Msg(X_ERROR, "%s: unknown EV_SYN code %d\n", local->name, event->code);
                break;
            case EV_REL:
                if(priv->debugLevel)
                    xf86Msg(X_ERROR, "EV_REL! %s: event type/code/value %d/%d/%d\n", local->name, event->type, event->code, event->value);
                int xaccel, yaccel;
                switch (event->code) {
                    case REL_X:
                        xaccel = 1;
                        if(priv->mouseAccel) {
                            if(abs(event->value) > 2) {
                                xaccel = 2;
                            } else if(abs(event->value) > 4) {
                                xaccel = 3;
                            } else if(abs(event->value) > 6) {
                                xaccel = 4;
                            }
                        }
                        x = x + (xaccel * priv->mouseSpeed * event->value);
                        if (x < priv->topX) {
                            x = 1 + priv->topX;
                        }
                        if (x > priv->bottomX) {
                            x = priv->bottomX;
                        }
                        break;
                    case REL_Y:
                        yaccel = 1;
                        if(priv->mouseAccel){
                            if (abs(event->value) > 2) {
                                yaccel = 2;
                            } else if(abs(event->value) > 4) {
                                yaccel = 3;
                            } else if(abs(event->value) > 6) {
                                yaccel = 4;
                            }
                        }
                        y = y + (yaccel * priv->mouseSpeed * event->value);
                        if (y < priv->topY) {
                            y = 1 + priv->topY;
                        }
                        if (y > priv->bottomY) {
                            y = priv->bottomY;
                        }
                        break;
                    case REL_WHEEL:
                        if (event->value < 0) {
                            buttons = set_bit(buttons,4,1);
                            delta = buttons ^ priv->oldButtons;
                            while (delta) {
                                id = ffs(delta);
                                delta &= ~(1 << (id-1));
                                if(priv->rotate90)
                                    xf86PostButtonEvent(local->dev, 1, id, (buttons&(1<<(id-1))), 0, 3, report_y, priv->screenX - report_x,report_z);
                                else
                                    xf86PostButtonEvent(local->dev, 1, id, (buttons&(1<<(id-1))), 0, 3, report_x, report_y,report_z);
                            }
                            priv->oldButtons = buttons;
                            buttons = set_bit(buttons,4,0);
                            delta = buttons ^ priv->oldButtons;
                            while (delta)
                            {
                                id = ffs(delta);
                                delta &= ~(1 << (id-1));
                                if(priv->rotate90)
                                    xf86PostButtonEvent(local->dev, 1, id, (buttons&(1<<(id-1))), 0, 3, report_y, priv->screenX - report_x,report_z);
                                else
                                    xf86PostButtonEvent(local->dev, 1, id, (buttons&(1<<(id-1))), 0, 3, report_x, report_y,report_z);
                            }
                        }
                        if (event->value > 0) {
                            buttons = set_bit(buttons,3,1);
                            delta = buttons ^ priv->oldButtons;
                            while (delta) {
                                id = ffs(delta);
                                delta &= ~(1 << (id-1));
                                if(priv->rotate90)
                                    xf86PostButtonEvent(local->dev, 1, id, (buttons&(1<<(id-1))), 0, 3, report_y, priv->screenX - report_x,report_z);
                                else
                                    xf86PostButtonEvent(local->dev, 1, id, (buttons&(1<<(id-1))), 0, 3, report_x, report_y,report_z);
                            }
                            priv->oldButtons = buttons;
                            buttons = set_bit(buttons,3,0);
                            delta = buttons ^ priv->oldButtons;
                            while (delta)
                            {
                                id = ffs(delta);
                                delta &= ~(1 << (id-1));
                                if(priv->rotate90)
                                    xf86PostButtonEvent(local->dev, 1, id, (buttons&(1<<(id-1))), 0, 3, report_y, priv->screenX - report_x,report_z);
                                else
                                    xf86PostButtonEvent(local->dev, 1, id, (buttons&(1<<(id-1))), 0, 3, report_x, report_y,report_z);
                            }
                        }
                        break;
                }
                break;
            case EV_ABS:
                switch (event->code) {
                    case ABS_X:
                        x = event->value;
                        if(priv->debugLevel)
                        xf86Msg(X_CONFIG, "ABS_X x is %d\n", x);
                        break;
                    case ABS_Z:
                        x = event->value;
                        if(priv->debugLevel)
                        xf86Msg(X_CONFIG, "ABS_Z x is %d\n", x);
                        break;
                    case ABS_RX:
                        y = event->value;
                        if(priv->debugLevel)
                            xf86Msg(X_CONFIG, "ABS_RX y is %d\n", y);
                        break;
                    case ABS_Y:
                        y = event->value;
                        if(priv->debugLevel)
                           xf86Msg(X_CONFIG, "ABS_Y y is %d\n", y);
                        break;

                    case ABS_PRESSURE:
                        if(    priv->activeButtons & WIZARDPEN_BTN_STYLUS
                            || priv->activeButtons & WIZARDPEN_BTN_STYLUS2)
                            break; /* Disable pressure events because this is actually part of a GPEN button event */

                        /* send button events if pressure crosses topZ threshold */
                        z = event->value;
                        if (z <= priv->topZ && priv->oldZ > priv->topZ) {
                            buttons = set_bit(buttons,0,0);
                            if(priv->debugLevel)
                                xf86Msg(X_CONFIG, "setting PEN UP event\n");
                        }
                        if ( priv->oldZ <= priv->topZ && z > priv->topZ ) {
                            buttons = set_bit(buttons,0,1);
                            if(priv->debugLevel)
                                xf86Msg(X_CONFIG, "setting PEN DOWN event\n");
                        }
                        if(priv->debugLevel)
                            xf86Msg(X_CONFIG, "ABS_PRESSURE z is %d (was %d)\n", z, priv->oldZ);
                        break;

                    case ABS_MISC:
                        break;

                }
                break; // EV_ABS

            case EV_KEY:
                switch (event->code) {
                    case BTN_TOOL_PEN:
                        prox = event->value;
                        if(priv->debugLevel)
                            xf86Msg(X_CONFIG, "BTN_TOOL_PEN prox is %d\n", prox);
                        break;

                    case BTN_LEFT:
                        buttons = set_bit(buttons,0,event->value);
                        if(priv->debugLevel)
                            xf86Msg(X_CONFIG, "NEW BTN_LEFT event is %d\n", event->value);
                        break;
                    case BTN_SIDE:
                        priv->activeButtons &= ~WIZARDPEN_BTN_SIDE;
                        if (event->value)
                            priv->activeButtons |= WIZARDPEN_BTN_SIDE;
                        // we detect a press ourself using the pressure in the code above
                        if(priv->debugLevel)
                            xf86Msg(X_CONFIG, "ignored BTN_SIDE event is %d\n", event->value);
                        break;

                    case BTN_EXTRA:
                        if(priv->debugLevel)
                            xf86Msg(X_CONFIG, "NEW BTN_EXTRA event is %d\n", event->value);
                        // NO BREAK
                    case BTN_MIDDLE:
                        buttons = set_bit(buttons,1,event->value);
                        if(priv->debugLevel)
                            xf86Msg(X_CONFIG, "BTN_MIDDLE event is %d\n", event->value);
                        break;

                    case BTN_FORWARD:
                        if(priv->debugLevel)
                            xf86Msg(X_CONFIG, "BTN_FORWARD event is %d\n", event->value);
                        // NO BREAK
                    case BTN_RIGHT:
                        buttons = set_bit(buttons,2,event->value);
                        if(priv->debugLevel)
                            xf86Msg(X_CONFIG, "BTN_RIGHT event is %d\n", event->value);
                        break;

                    /*Code for handling the buttons on GPEN styluses
                    * Pressure MUST be disabled otherwise left-click events will be sent.
                    F509:
                    Top button: BTN_STYLUS = 1, BTN_STYLUS = 0
                    Bottom button: BTN_TOUCH =1, BTN_STYLUS = 1, BTN_TOUCH = 0, BTN_STYLUS = 0

                    560:
                    1st side button: BTN_STYLUS =1, BTN_STYLUS = 0
                    2nd side button: BTN_STYLUS2 = 1, BTN_STYLUS2 = 0
                    */
                    case BTN_TOUCH:
                        if(priv->debugLevel)
                            xf86Msg(X_CONFIG, "BTN_TOUCH event is %d\n", event->value);
                        priv->activeButtons ^= WIZARDPEN_BTN_TOUCH; //Toggle event
                        if(!event->value)
                        {
                            buttons = set_bit(buttons, 2, 0); /* try to clear events to prevent some becoming 'locked' in the active state */
                            buttons = set_bit(buttons, 1, 0);
                        }
                        break;
                    case BTN_STYLUS:
                        if(priv->debugLevel)
                        xf86Msg(X_CONFIG, "BTN_STYLUS event is %d\n", event->value);
                        priv->activeButtons ^= WIZARDPEN_BTN_STYLUS;

                        buttons = set_bit(buttons, 1, 0); //HACK, to avoid having to do conditional sections depending on tablet type

                        if(priv->activeButtons & WIZARDPEN_BTN_TOUCH)
                            buttons = set_bit(buttons, 2, event->value);
                        else
                        {
                            buttons = set_bit(buttons, 1, event->value);
                            if(!event->value)
                                buttons = set_bit(buttons, 2, 0); /* Clear right click event as well, due to clearing order */
                        }
                        break;
                    case BTN_STYLUS2:
                        if (priv->debugLevel)
                            xf86Msg(X_CONFIG, "BTN_STYLUS2 event is %d\n", event->value);
                        priv->activeButtons ^= WIZARDPEN_BTN_STYLUS2;
                        buttons = set_bit(buttons, 2, event->value);
                        break;
                    default:
                        if (priv->debugLevel)
                            xf86Msg(X_CONFIG, "Unhandled event %d is %d\n", event->code, event->value);
                        break;
                }
                break; // EV_KEY
            case EV_MSC:
                break;
            default:
                xf86Msg(X_ERROR, "%s: unknown event type/code/value %d/%d/%d\n", local->name, event->type, event->code, event->value);
        } /* switch event->type */

        /* If pressure is below threshold, set pressure and reported pressure to 0 */
        if (z <= priv->topZ) {
            report_z = 0;
        } else if(z >= priv->bottomZ) {
            report_z = 1023;
        } else { /* Otherwise scale down by min pressure val */
            report_z = z - priv->topZ;
            report_z = 1023.0 * ( ((float)report_z) / (priv->bottomZ - priv->topZ));
        }

        if (priv->reportScreen)
        {
            // we now convert x and y to screen coordinates
            if (x>priv->topX && x<priv->bottomX)
                report_x = (x-priv->topX) * priv->screenX / (priv->bottomX-priv->topX);
            else{
                if(x<priv->topX)
                    report_x = 0;
                else
                    report_x = priv->screenX;
            }
            if (y>priv->topY && y<priv->bottomY)
                report_y = (y-priv->topY) * priv->screenY / (priv->bottomY-priv->topY);
            else {
                if(y<priv->topY)
                    report_y = 0;
                else
                    report_y = priv->screenY;
            }
        }
        else
        {
            // make sure it is inside the active area
            report_x = max(priv->topX, min(x, priv->bottomX));
            report_y = max(priv->topY, min(y, priv->bottomY));
        }

        // rotatate if specified
        if (priv->rotate90) {
            int tempval;
            if (report_y < 0)
                report_y = 0;
            if (report_x < 0)
                report_x = 0;
            tempval = report_x;
            report_x = (priv->screenY - report_y);
            // FIXME: What is this for? Does look suspiciously resolution dependent
            if (report_x > 1278)
                report_x = 1278;
            report_y = tempval;
        }

        report_buttons = buttons;
        // Tablet PC buttons
        if (priv->tpcButton)
        {
            if (   buttons & 1                              // Pressure click
                || priv->activeButtons & WIZARDPEN_BTN_SIDE // Trust TB-6300 no longer reports pressure when one of the buttons is pressed
                )
            {
                if (buttons & ~1) // any other buttons pressed?
                    report_buttons &= ~1; // ignore the left click
            }
            else
                report_buttons &= 1;  // suppress the other buttons
        }

        // We ALWAYS send button events right after we notice them
        // The pen might get stuck in pressed down state otherwise
        if (priv->oldReportedButtons != report_buttons)
        {
            int delta = report_buttons ^ priv->oldReportedButtons;

            while (delta)
            {
                int id = ffs(delta);
                delta &= ~(1 << (id-1)); // clear it
                xf86PostButtonEvent(local->dev, 1, id, (report_buttons&(1<<(id-1))), 0, 3, report_x, report_y, report_z);
            }
        }

        priv->oldButtons = buttons;
        priv->oldReportedButtons = report_buttons;
        priv->oldX = x;
        priv->oldY = y;
        priv->oldZ = z;

        /* Linux Kernel 2.6.x sends EV_SYN/SYN_REPORT as an event terminator,
         * whereas 2.4.x sends EV_ABS/ABS_MISC. We have to support both.
         * It seems that these don't always arrive at the end of a
         * series of events.
         */
        if ((event->type == EV_SYN && event->code == SYN_REPORT) ||
            (event->type == EV_ABS && event->code == ABS_MISC))
        {
            if (!(priv->oldProximity))
                if (!is_core_pointer)
                {
                    xf86PostProximityEvent(local->dev, 1, 0, 3 ,
                        report_x, report_y, report_z);
                }
            priv->oldProximity = prox;

            xf86PostMotionEvent(local->dev, 1, 0, 3, report_x, report_y, report_z);
        }
    }
}
#endif // #ifdef LINUX_INPUT

#define WriteString(str)\
    XisbWrite (priv->buffer, (unsigned char *)(str), strlen(str))

static Bool
QueryHardware (WizardPenPrivatePtr priv)
{
    /* Reset */
    WriteString(WIZARDPEN_RESET);

    /* Wait */
    milisleep (250);

    /* Prompt Mode in order to not be disturbed */
    WriteString(WIZARDPEN_PROMPT_MODE);

    /* Flush */
    while (XisbRead(priv->buffer) >= 0);

    /* Ask for Config packet*/
    WriteString(WIZARDPEN_CONFIG);

    /* Read the packet */
    XisbBlockDuration (priv->buffer, 1000000);
    NewPacket (priv);

    if ((WizardPenGetPacket (priv) == Success))
    {
        priv->maxX = (int)priv->packet[1] + ((int)priv->packet[2] << 7);
        priv->maxY = (int)priv->packet[3] + ((int)priv->packet[4] << 7);
        priv->maxZ = 512; // FIXME: some tablets go to 1024
        xf86Msg(X_PROBED, "Wizardpen Tablet MaxX:%d MaxY:%d\n", priv->maxX, priv->maxY);
    }
    else
        return !Success;

    return Success;
}

#define BITS_PER_LONG (sizeof(long) * 8)
#define NBITS(x) ((((x)-1)/BITS_PER_LONG)+1)
#define test_bit(bit, array)    ((array[LONG(bit)] >> OFF(bit)) & 1)
#define OFF(x)  ((x)%BITS_PER_LONG)
#define LONG(x) ((x)/BITS_PER_LONG)

#ifdef LINUX_INPUT
static Bool
USBQueryHardware (InputInfoPtr local)
{
    WizardPenPrivatePtr priv = (WizardPenPrivatePtr) local->private;
    unsigned long   bit[EV_MAX][NBITS(KEY_MAX)];
    int         i, j;
    int         abs[5];
    char        name[256] = "Unknown";
    int         aX, bX, aY, bY;

    ioctl(local->fd, EVIOCGNAME(sizeof(name)), name);
    xf86MsgVerb(X_PROBED, 4, "Kernel Input device name: \"%s\"\n", name);

    memset(bit, 0, sizeof(bit));
    ioctl(local->fd, EVIOCGBIT(0, EV_MAX), bit[0]);

    priv->maxX = 0;
    priv->maxY = 0;
    priv->maxZ = 0;
    for (i = 0; i < EV_MAX; i++)
        if (test_bit(i, bit[0])) {
            ioctl(local->fd, EVIOCGBIT(i, KEY_MAX), bit[i]);
            for (j = 0; j < KEY_MAX; j++)
                if (test_bit(j, bit[i])) {
                    if (i == EV_ABS) {
                        ioctl(local->fd, EVIOCGABS(j), abs);
                        switch (j) {
                            case ABS_X:
                                aX = abs[2];
                                break;
                            case ABS_Z:
                                bX = abs[2];
                                break;
                            case ABS_Y:
                                aY = abs[2];
                                break;
                            case ABS_RX:
                                bY = abs[2];
                                break;
                            case ABS_PRESSURE:
                                priv->maxZ = abs[2];
                                break;
                        }
                    }
                }
        }
    
    // FIXME: do we need to check this or is it always the same?
    if (aX + aY > bX + bY)
    {
        priv->maxX = aX;
        priv->maxY = aY;
        if (bY > 0)
            priv->aspectRatio = (float)bX/bY;
    }
    else
    {
        priv->maxX = bX;
        priv->maxY = bY;
        if (aY > 0)
            priv->aspectRatio = (float)aX/aY;
    }

    // Should this be done here?
    if(!priv->bottomX)
        priv->bottomX = priv->maxX;
    if(!priv->bottomY)
        priv->bottomY = priv->maxY;
    if(!priv->bottomZ)
        priv->bottomZ = priv->maxZ;

    xf86Msg(X_PROBED, "%s: MaxX:%d MaxY:%d MaxZ:%d\n", local->name, priv->maxX, priv->maxY, priv->maxZ);
    xf86Msg(X_PROBED, "%s: aspect ratio:%.2f:1\n",       local->name, priv->aspectRatio);
    return Success;
}
#endif // #ifdef LINUX_INPUT

static void
NewPacket (WizardPenPrivatePtr priv)
{
    priv->packeti = 0;
}

static Bool
WizardPenGetPacket (WizardPenPrivatePtr priv)
{
    int count = 0;
    int c = 0;

    while((c = XisbRead(priv->buffer)) >= 0 )
    {

        /*
         * fail after 500 bytes so the server doesn't hang forever if a
         * device sends bad data.
         */
        if (count++ > 500)
        {
            NewPacket (priv);
            return !Success;
        }

        if (c & PHASING_BIT)
        {
            NewPacket(priv);

            XisbBlockDuration (priv->buffer, 10000);
            priv->packet[priv->packeti++] = c;
            count = WIZARDPEN_PACKET_SIZE - 1;
            while (count-- && (c = XisbRead(priv->buffer)) >= 0)
            {
                priv->packet[priv->packeti++] = c;
            }
            XisbBlockDuration (priv->buffer, 0);
            if(c > 0)
                return Success;
        }
    }
    return !Success;
}
